from pydantic import BaseModel, Field

from schorle.element import Element, html
from schorle.theme import Theme


class Document(BaseModel):
    title: str = "Schorle"
    theme: Theme = Theme.DARK
    with_dev_meta: bool = False
    extra_assets: list | None = None
    lang: str = "en"
    with_tailwind: bool = True
    with_daisyui: bool = True
    daisyui_version: str = "4.7.2"
    body_attributes: dict[str, str] | None = Field(default_factory=lambda: {"hx-ext": "morph, lucide"})

    def _base(self):
        with html(lang=self.lang, theme=self.theme, **{"data-theme": self.theme}) as _html:
            with _html.head() as _head:
                _head.meta(charset="utf-8")
                _head.meta(name="viewport", content="width=device-width, initial-scale=1.0")
                _head.title().text(self.title)
                if self.with_dev_meta:
                    _head.meta(name="schorle-dev", content="true")

                _head.link(href="/favicon.svg", rel="icon", type="image/svg+xml")
                if self.with_tailwind:
                    _head.script(src="https://cdn.tailwindcss.com")
                if self.with_daisyui:
                    _head.link(
                        href=f"https://cdn.jsdelivr.net/npm/daisyui@{self.daisyui_version}/dist/full.min.css",
                        rel="stylesheet",
                        type="text/css",
                    )
                _head.script(src="https://unpkg.com/htmx.org@1.9.10/dist/htmx.min.js")
                _head.script(src="https://unpkg.com/idiomorph@0.3.0")
                _head.script(src="https://unpkg.com/idiomorph@0.3.0/dist/idiomorph-ext.min.js")
                _head.script(src="https://unpkg.com/hyperscript.org@0.9.12")
                _head.script(src="/_schorle/assets/bundle.js", crossorigin="anonymous", **{"defer": ""})

                if self.extra_assets:
                    for asset in self.extra_assets:
                        _head.append(asset)

            return _html

    def include(self, payload: Element) -> Element:
        _html = self._base()
        _html.body(**self.body_attributes).append(payload)
        return _html



from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import Any, Union

from pydantic import BaseModel, Field

RawClassesPayload = Union[str, list[str], tuple[str, ...], "Classes", None]


def parse_args(*args: RawClassesPayload) -> list[str]:
    container = []
    for arg in args:
        if isinstance(arg, str):
            container.append(arg)
        elif isinstance(arg, (list, tuple)):
            container.extend(arg)
        elif isinstance(arg, Classes):
            container.extend(arg.value)
        elif arg is None:
            pass
        else:
            msg = f"Invalid type: {type(arg)} of {arg} for Classes"
            raise TypeError(msg)
    return container


class Classes(BaseModel):
    value: list[str] = Field(default_factory=list)

    def __init__(self, *args: RawClassesPayload):
        super().__init__()
        self.value = parse_args(*args)

    def append(self, *args: RawClassesPayload):
        new_value = self.value + parse_args(*args)
        self.value = new_value
        return self

    def remove(self, *args: RawClassesPayload):
        new_value = [x for x in self.value if x not in parse_args(*args)]
        self.value = new_value
        return self

    def toggle(self, class_name: str):
        if class_name in self.value:
            self.remove(class_name)
        else:
            self.append(class_name)
        return self

    def render(self) -> str:
        return "" if not self.value else " ".join(sorted(set(self.value))).strip()

    def __str__(self) -> str:
        return self.render()

    def __repr__(self) -> str:
        return f"Classes({self.render()})"


@dataclass
class On:
    trigger: str
    callback: Any


class HTTPMethod(str, Enum):
    POST = "POST"
    GET = "GET"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"


@dataclass
class Action:
    method: HTTPMethod
    url: str


def post(url: str):
    return Action(method=HTTPMethod.POST, url=url)


def get(url: str):
    return Action(method=HTTPMethod.GET, url=url)


def put(url: str):
    return Action(method=HTTPMethod.PUT, url=url)


def delete(url: str):
    return Action(method=HTTPMethod.DELETE, url=url)


def patch(url: str):
    return Action(method=HTTPMethod.PATCH, url=url)


class Swap(str, Enum):
    morph = "morph"
    after_end = "afterend"
    before_begin = "beforebegin"
    before_end = "beforeend"
    outer_html = "outerHtml"
    inner_html = "innerHtml"


@dataclass
class Reactive:
    action: Action
    target: str
    swap: str | Swap = Swap.morph
    trigger: str | None = None

    def render(self) -> dict[str, str]:
        _rendered = {}

        if self.trigger:
            _rendered["hx-trigger"] = self.trigger

        _rendered.update(
            {
                "hx-swap": self.swap,
                "hx-target": self.target,
                "hx-" + self.action.method.lower(): self.action.url,
            }
        )
        return _rendered



from schorle.attrs import Classes
from schorle.element import div


def test_el():
    with div() as d:
        d.div().text("hey")

    assert d.render(pretty_print=False) == "<div><div>hey</div></div>"

    with div() as d:
        d.div().text("hey")
        d.div().text("ho")

    assert d.render(pretty_print=False) == "<div><div>hey</div><div>ho</div></div>"


def test_with_attrs():
    with div(element_id="level1") as d:
        d.div(element_id="level2").text("hey")

    assert d.render(pretty_print=False) == '<div id="level1"><div id="level2">hey</div></div>'

    with div(element_id="level1", classes=Classes("bg-red-500")) as d:
        d.div(element_id="level2").text("hey")

    assert d.render(pretty_print=False) == '<div id="level1" class="bg-red-500"><div id="level2">hey</div></div>'



from __future__ import annotations

from lxml import etree
from pydantic import BaseModel, PrivateAttr

from schorle.attrs import Classes, Reactive
from schorle.tags import HTMLTag
from schorle.types import LXMLElement


class ElementPrototype(BaseModel):
    tag: HTMLTag
    element_id: str | None = None
    classes: Classes | None = None
    style: dict[str, str] | None = None
    attrs: dict[str, str] | None = None
    reactive: Reactive | None = None
    _children: list[ElementPrototype] = PrivateAttr(default_factory=list)
    _text: str | None = PrivateAttr(default=None)

    def append(self, *children: ElementPrototype):
        for child in children:
            self._children.append(child)
        return self

    def get_children(self) -> list[ElementPrototype]:
        return self._children

    def to_lxml(self) -> LXMLElement:
        _element = etree.Element(self.tag.value)
        for k, v in self.get_lxml_element_attrs().items():
            _element.set(k, v)

        if self._text:
            _element.text = self._text
        return _element

    def get_lxml_element_attrs(self) -> dict[str, str]:
        _attributes = self.attrs or {}
        if self.element_id:
            _attributes["id"] = self.element_id

        if self.classes:
            _attributes["class"] = self.classes.render()

        if self.style:
            _attributes["style"] = ";".join([f"{k}:{v}" for k, v in self.style.items()])

        if self.reactive:
            _attributes.update(self.reactive.render())

        return _attributes

    def text(self, text: str):
        self._text = text
        return self



import mimetypes
from importlib.resources import files
from pathlib import Path

from fastapi import FastAPI
from starlette.responses import FileResponse
from starlette.types import Receive, Scope, Send

from schorle.document import Document
from schorle.theme import Theme

ASSETS_PATH = Path(str(files("schorle"))) / Path("assets")


def favicon() -> FileResponse:
    favicon_path = ASSETS_PATH / "favicon.svg"
    return FileResponse(favicon_path, media_type="image/svg+xml")


def assets(file_name: str) -> FileResponse:
    file_path = ASSETS_PATH / file_name
    mime_type, _ = mimetypes.guess_type(file_path)
    return FileResponse(file_path, media_type=mime_type)


class Schorle:
    def __init__(self, theme: Theme = Theme.DARK, lang: str = "en", extra_assets: list | None = None) -> None:
        self.backend = FastAPI()
        self.backend.get("/_schorle/assets/{file_name:path}", response_class=FileResponse)(assets)
        self.backend.get("/favicon.svg", response_class=FileResponse)(favicon)
        self.theme: Theme = theme
        self.extra_assets = extra_assets
        self.lang = lang

    async def __call__(self, scope: Scope, receive: Receive, send: Send):
        """
        This method is called by uvicorn when the server is started.
        """
        await self.backend(scope=scope, receive=receive, send=send)

    @property
    def doc(self) -> Document:
        return Document(theme=self.theme, lang=self.lang, extra_assets=self.extra_assets)



from pathlib import Path

from starlette.responses import FileResponse, JSONResponse

from schorle.app import Schorle
from schorle.attrs import Classes
from schorle.element import a, div, link
from schorle.icon import icon

app = Schorle(
    extra_assets=[
        link(href="https://fonts.googleapis.com/css?family=Space+Mono", rel="stylesheet", type="text/css"),
    ]
)


def LinkWithIcon(href: str, icon_name: str, additional_text: str) -> a:
    with a(href=href, classes=Classes("btn btn-primary font-normal w-42")) as this:
        this.text(additional_text)
        this >> icon(icon_name)
    return this


LINKS = [
    LinkWithIcon("https://github.com/renardeinside/schorle", "github", "GitHub"),
    LinkWithIcon("https://github.com/renardeinside/schorle/tree/main/examples", "code", "Examples"),
    LinkWithIcon(
        "https://medium.com/@polarpersonal/schorle-testing-the-waters-with-a-python-server-driven-ui-kit-053f85ee6574",
        "book-marked",
        "Concepts",
    ),
]


@app.backend.get("/logo", response_class=FileResponse)
def logo():
    return FileResponse(path=Path(__file__).parent.parent / Path("raw/with_text.svg"), media_type="image/svg+xml")


@app.backend.get("/")
def landing_page():
    with div(
        classes=Classes("flex flex-col justify-center items-center h-screen"), style={"font-family": "Space Mono"}
    ) as mv:
        mv.div(classes=Classes("max-w-screen-md flex flex-col justify-center items-center space-y-4")).img(
            src="/logo",
            alt="Schorle logo",
            media_type="image/svg+xml",
            classes=Classes("w-3/4"),
        )
        mv.p(classes=Classes("text-2xl m-4 p-4 w-5/6 text-center")).text(
            "Pythonic Server-Driven UI Kit for building modern apps."
        )

        mv.div(classes=Classes("flex flex-col space-y-4 space-x-0 md:flex-row md:space-x-4 md:space-y-0")).append(
            *LINKS
        )
    return app.doc.include(mv).to_response()


@app.backend.get("/health", response_class=JSONResponse)
def health():
    return {"status": "ok"}



from __future__ import annotations

from functools import partial
from http import HTTPStatus
from typing import Any, Callable

from lxml import etree
from starlette.responses import HTMLResponse

from schorle.attrs import Classes, Reactive
from schorle.prototypes import ElementPrototype
from schorle.tags import HTMLTag
from schorle.types import LXMLElement
from schorle.utils import fix_self_closing_tags


class Element(ElementPrototype):
    post_callback: Callable[[Element], Any] | None = None

    def __init__(self, **data):
        super().__init__(**data)
        if self.post_callback:
            self.post_callback(self)

    def render(self, *, pretty_print: bool = True) -> str:
        _composed = self._compose(self)
        return etree.tostring(_composed, pretty_print=pretty_print, encoding="utf-8").decode("utf-8")

    def to_response(self, status_code: HTTPStatus = HTTPStatus.OK, *args, **kwargs) -> HTMLResponse:
        return HTMLResponse(self.render(), status_code.value, *args, **kwargs)

    def _compose(self, element: ElementPrototype) -> LXMLElement:
        _element = element.to_lxml()
        for child in element.get_children():
            _element.append(self._compose(child))
        fix_self_closing_tags(_element)
        return _element

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def __getattr__(self, name):
        if name.upper() in HTMLTag.__members__:

            def post_callback(element: Element):
                self.append(element)

            _factory = element_function_factory(HTMLTag[name.upper()])
            return partial(_factory, post_callback=post_callback)
        return super().__getattr__(name)

    def __rshift__(self, other):
        self.append(other)
        return self


def element_function_factory(tag: HTMLTag):
    def func(
        classes: Classes | None = None,
        element_id: str | None = None,
        style: dict[str, str] | None = None,
        attrs: dict[str, str] | None = None,
        reactive: Reactive | None = None,
        hsx: str | None = None,
        post_callback: Callable[[Element], Any] | None = None,
        **attributes,
    ):
        combined_attrs = {**attributes, **(attrs or {})}
        if hsx:
            combined_attrs["data-script"] = str(hsx)
        return Element(
            tag=tag,
            element_id=element_id,
            classes=classes,
            style=style,
            reactive=reactive,
            attrs=combined_attrs,
            post_callback=post_callback,
        )

    return func


div = element_function_factory(HTMLTag.DIV)
span = element_function_factory(HTMLTag.SPAN)
button = element_function_factory(HTMLTag.BUTTON)
input_ = element_function_factory(HTMLTag.INPUT)
a = element_function_factory(HTMLTag.A)
img = element_function_factory(HTMLTag.IMG)
script = element_function_factory(HTMLTag.SCRIPT)
link = element_function_factory(HTMLTag.LINK)
meta = element_function_factory(HTMLTag.META)
title = element_function_factory(HTMLTag.TITLE)
head = element_function_factory(HTMLTag.HEAD)
body = element_function_factory(HTMLTag.BODY)
html = element_function_factory(HTMLTag.HTML)
p = element_function_factory(HTMLTag.P)
h1 = element_function_factory(HTMLTag.H1)
h2 = element_function_factory(HTMLTag.H2)
h3 = element_function_factory(HTMLTag.H3)
h4 = element_function_factory(HTMLTag.H4)
h5 = element_function_factory(HTMLTag.H5)
h6 = element_function_factory(HTMLTag.H6)
ul = element_function_factory(HTMLTag.UL)
ol = element_function_factory(HTMLTag.OL)
li = element_function_factory(HTMLTag.LI)
table = element_function_factory(HTMLTag.TABLE)
tr = element_function_factory(HTMLTag.TR)
td = element_function_factory(HTMLTag.TD)
th = element_function_factory(HTMLTag.TH)
form = element_function_factory(HTMLTag.FORM)
footer = element_function_factory(HTMLTag.FOOTER)
icon = element_function_factory(HTMLTag.ICON)



from __future__ import annotations

from typing import Annotated

from fastapi import Form
from loguru import logger

from schorle.app import Schorle
from schorle.attrs import Classes, Reactive, Swap, delete, post
from schorle.element import button, div, form, input_, span
from schorle.icon import icon
from schorle.theme import Theme

app = Schorle(theme=Theme.DARK)


def task_view(task: str):
    with div(Classes("flex flex-row justify-between items-center w-full")) as this:
        this >> span(Classes("text-lg")).text(task)
        this >> button(
            Classes("btn btn-error btn-square"), reactive=Reactive(delete(f"/delete?task={task}"), "closest div")
        ).append(icon("trash"))

    return this


def text_input():
    return input_(
        placeholder="What needs to be done?",
        name="task",
        classes=Classes("input input-primary w-96"),
        hsx="""
        on keyup[event.key == 'Escape']
            set my.value to ''
            my.blur()
        on keyup
            if my.value.length > 0
                remove .btn-disabled from #add-button
            else
                add .btn-disabled to #add-button
        """,
    )


def input_component():
    with form(
        Classes("flex flex-row justify-center items-center w-96 space-x-2"),
        reactive=Reactive(post("/add"), "#tasks", Swap.before_end),
        hsx="""
        on htmx:afterRequest
            my.reset()
            add .btn-disabled to #add-button
        """,
    ) as this:
        this >> text_input()
        this >> button(
            Classes("btn btn-primary btn-square btn-disabled"),
            element_id="add-button",
        ).append(icon("plus"))

    return this


def tasks_view():
    this = div(Classes("flex flex-col justify-center items-center w-96 space-y-2 mt-2"), element_id="tasks")
    for task in app.backend.state.tasks:
        this >> task_view(task)
    return this


def main_view():
    with div(Classes("flex flex-col justify-center items-center h-screen")) as mv:
        mv >> input_component()
        mv >> tasks_view()
    return mv


@app.backend.get("/")
def home():
    app.backend.state.tasks = ["task 1", "task 2", "task 3"]
    return app.doc.include(main_view()).to_response()


@app.backend.post("/add")
def add_task(task: Annotated[str, Form()]):
    logger.info(f"Adding task: {task}")
    app.backend.state.tasks.append(task)
    return task_view(task).to_response()


@app.backend.delete("/delete")
def delete_task(task: str):
    logger.info(f"Deleting task: {task}")
    app.backend.state.tasks.remove(task)



from __future__ import annotations

import sys
from enum import Enum

from starlette.responses import HTMLResponse

from schorle.types import LXMLElement


class RunningMode(str, Enum):
    DEV = "dev"
    PRODUCTION = "production"


def get_running_mode() -> RunningMode:
    # if uvicorn and --reload are in sys.argv, we are running in uvicorn dev mode
    _joined_argv = " ".join(sys.argv)
    if "uvicorn" in _joined_argv and "--reload" in _joined_argv:
        return RunningMode.DEV
    else:
        return RunningMode.PRODUCTION


def empty():
    return HTMLResponse(content="", status_code=200)


def fix_self_closing_tags(element: LXMLElement) -> None:
    for _tag in element.iter():
        if _tag.tag in ["script", "link", "i", "span"]:
            if _tag.text is None:
                _tag.text = ""



from __future__ import annotations

from schorle.app import Schorle
from schorle.attrs import Classes
from schorle.element import div
from schorle.theme import Theme

app = Schorle(theme=Theme.DARK)


def main_view():
    with div(Classes("flex flex-col items-center justify-center h-screen")) as mv:
        mv.button(
            Classes("btn btn-primary"),
            hsx="""
            init
                set my.count to 0
            on click
                set my.count to my.count + 1
                put `Clicked ${my.count} times!` into me
            """,
        ).text("Click me!")
    return mv


@app.backend.get("/")
def home():
    return app.doc.include(main_view()).to_response()



import textwrap

from lxml import etree

from schorle.document import Document


def test_empty_doc():
    _title = "Test Document"
    doc = Document(title=_title)
    lnk = f"https://cdn.jsdelivr.net/npm/daisyui@{doc.daisyui_version}/dist/full.min.css"
    result = etree.tostring(doc._base()._compose(doc._base()), pretty_print=True).decode()
    expected = textwrap.dedent(
        f"""\
        <html lang="en" theme="dark" data-theme="dark">
          <head>
            <meta charset="utf-8"/>
            <meta name="viewport" content="width=device-width, initial-scale=1.0"/>
            <title>{_title}</title>
            <link href="/favicon.svg" rel="icon" type="image/svg+xml"></link>
            <script src="https://cdn.tailwindcss.com"></script>
            <link href="{lnk}" rel="stylesheet" type="text/css"></link>
            <script src="https://unpkg.com/htmx.org@1.9.10/dist/htmx.min.js"></script>
            <script src="https://unpkg.com/idiomorph@0.3.0"></script>
            <script src="https://unpkg.com/idiomorph@0.3.0/dist/idiomorph-ext.min.js"></script>
            <script src="https://unpkg.com/hyperscript.org@0.9.12"></script>
            <script src="/_schorle/assets/bundle.js" crossorigin="anonymous" defer=""></script>
          </head>
        </html>
    """
    )
    assert result == expected
